@@ -9,6 +9,7 @@ use cranelift_codegen::ir::{
 };
 use log::trace;
 use smallvec::{smallvec, SmallVec};
+use std::convert::TryFrom;
 use std::ops::RangeFrom;
 use thiserror::Error;
 
@@ -71,6 +72,18 @@ where
     // Indicate that the result of a step is to assign a single value to an instruction's results.
     let assign = |value: V| ControlFlow::Assign(smallvec![value]);
 
+    // Indicate that the result of a step is to assign multiple values to an instruction's results.
+    let assign_multiple = |values: &[V]| ControlFlow::Assign(SmallVec::from(values));
+
+    // Similar to `assign` but converts some errors into traps
+    let assign_or_trap = |value: ValueResult<V>| match value {
+        Ok(v) => Ok(assign(v)),
+        Err(ValueError::IntegerDivisionByZero) => Ok(ControlFlow::Trap(CraneliftTrap::User(
+            TrapCode::IntegerDivisionByZero,
+        ))),
+        Err(e) => Err(e),
+    };
+
     // Interpret a binary instruction with the given `op`, assigning the resulting value to the
     // instruction's results.
     let binary = |op: fn(V, V) -> ValueResult<V>,
@@ -78,18 +91,24 @@ where
                   right: V|
      -> ValueResult<ControlFlow<V>> { Ok(assign(op(left, right)?)) };
 
-    // Same as `binary`, but converts the values to their unsigned form before the operation and
-    // back to signed form afterwards. Since Cranelift types have no notion of signedness, this
-    // enables operations that depend on sign.
-    let binary_unsigned =
+    // Similar to `binary` but converts select `ValueError`'s into trap `ControlFlow`'s
+    let binary_can_trap = |op: fn(V, V) -> ValueResult<V>,
+                           left: V,
+                           right: V|
+     -> ValueResult<ControlFlow<V>> { assign_or_trap(op(left, right)) };
+
+    // Same as `binary_can_trap`, but converts the values to their unsigned form before the
+    // operation and back to signed form afterwards. Since Cranelift types have no notion of
+    // signedness, this enables operations that depend on sign.
+    let binary_unsigned_can_trap =
         |op: fn(V, V) -> ValueResult<V>, left: V, right: V| -> ValueResult<ControlFlow<V>> {
-            Ok(assign(
+            assign_or_trap(
                 op(
                     left.convert(ValueConversionKind::ToUnsigned)?,
                     right.convert(ValueConversionKind::ToUnsigned)?,
-                )?
-                .convert(ValueConversionKind::ToSigned)?,
-            ))
+                )
+                .and_then(|v| v.convert(ValueConversionKind::ToSigned)),
+            )
         };
 
     // Choose whether to assign `left` or `right` to the instruction's result based on a `condition`.
@@ -102,8 +121,17 @@ where
 
     // Based on `condition`, indicate where to continue the control flow.
     let branch_when = |condition: bool| -> Result<ControlFlow<V>, StepError> {
+        let branch_args = match inst {
+            InstructionData::Jump { .. } => args_range(0..),
+            InstructionData::BranchInt { .. }
+            | InstructionData::BranchFloat { .. }
+            | InstructionData::Branch { .. } => args_range(1..),
+            InstructionData::BranchIcmp { .. } => args_range(2..),
+            _ => panic!("Unrecognized branch inst: {:?}", inst),
+        }?;
+
         Ok(if condition {
-            ControlFlow::ContinueAt(branch(), args_range(1..)?)
+            ControlFlow::ContinueAt(branch(), branch_args)
         } else {
             ControlFlow::Continue
         })
@@ -122,7 +150,7 @@ where
     };
 
     // Helper for summing a sequence of values.
-    fn sum<V: Value>(head: V, tail: SmallVec<[V; 1]>) -> ValueResult<i64> {
+    fn sum<V: Value>(head: V, tail: SmallVec<[V; 1]>) -> ValueResult<i128> {
         let mut acc = head;
         for t in tail {
             acc = Value::add(acc, t)?;
@@ -133,15 +161,38 @@ where
     // Interpret a Cranelift instruction.
     Ok(match inst.opcode() {
         Opcode::Jump | Opcode::Fallthrough => ControlFlow::ContinueAt(branch(), args()?),
-        Opcode::Brz => branch_when(!arg(0)?.into_bool()?)?,
-        Opcode::Brnz => branch_when(arg(0)?.into_bool()?)?,
-        Opcode::BrIcmp => branch_when(icmp(inst.cond_code().unwrap(), &arg(1)?, &arg(2)?)?)?,
+        Opcode::Brz => branch_when(
+            !arg(0)?
+                .convert(ValueConversionKind::ToBoolean)?
+                .into_bool()?,
+        )?,
+        Opcode::Brnz => branch_when(
+            arg(0)?
+                .convert(ValueConversionKind::ToBoolean)?
+                .into_bool()?,
+        )?,
+        Opcode::BrIcmp => branch_when(icmp(inst.cond_code().unwrap(), &arg(0)?, &arg(1)?)?)?,
         Opcode::Brif => branch_when(state.has_iflag(inst.cond_code().unwrap()))?,
         Opcode::Brff => branch_when(state.has_fflag(inst.fp_cond_code().unwrap()))?,
-        Opcode::BrTable => unimplemented!("BrTable"),
-        Opcode::JumpTableEntry => unimplemented!("JumpTableEntry"),
-        Opcode::JumpTableBase => unimplemented!("JumpTableBase"),
-        Opcode::IndirectJumpTableBr => unimplemented!("IndirectJumpTableBr"),
+        Opcode::BrTable => {
+            if let InstructionData::BranchTable {
+                table, destination, ..
+            } = inst
+            {
+                let jt_data = &state.get_current_function().jump_tables[table];
+
+                // Convert to usize to remove negative indexes from the following operations
+                let jump_target = usize::try_from(arg(0)?.into_int()?)
+                    .ok()
+                    .and_then(|i| jt_data.as_slice().get(i))
+                    .copied()
+                    .unwrap_or(destination);
+
+                ControlFlow::ContinueAt(jump_target, SmallVec::new())
+            } else {
+                unreachable!()
+            }
+        }
         Opcode::Trap => ControlFlow::Trap(CraneliftTrap::User(trap_code())),
         Opcode::Debugtrap => ControlFlow::Trap(CraneliftTrap::Debug),
         Opcode::ResumableTrap => ControlFlow::Trap(CraneliftTrap::Resumable),
@@ -392,16 +443,16 @@ where
         Opcode::Imul => binary(Value::mul, arg(0)?, arg(1)?)?,
         Opcode::Umulhi => unimplemented!("Umulhi"),
         Opcode::Smulhi => unimplemented!("Smulhi"),
-        Opcode::Udiv => binary_unsigned(Value::div, arg(0)?, arg(1)?)?,
-        Opcode::Sdiv => binary(Value::div, arg(0)?, arg(1)?)?,
-        Opcode::Urem => binary_unsigned(Value::rem, arg(0)?, arg(1)?)?,
-        Opcode::Srem => binary(Value::rem, arg(0)?, arg(1)?)?,
+        Opcode::Udiv => binary_unsigned_can_trap(Value::div, arg(0)?, arg(1)?)?,
+        Opcode::Sdiv => binary_can_trap(Value::div, arg(0)?, arg(1)?)?,
+        Opcode::Urem => binary_unsigned_can_trap(Value::rem, arg(0)?, arg(1)?)?,
+        Opcode::Srem => binary_can_trap(Value::rem, arg(0)?, arg(1)?)?,
         Opcode::IaddImm => binary(Value::add, arg(0)?, imm_as_ctrl_ty()?)?,
         Opcode::ImulImm => binary(Value::mul, arg(0)?, imm_as_ctrl_ty()?)?,
-        Opcode::UdivImm => binary_unsigned(Value::div, arg(0)?, imm())?,
-        Opcode::SdivImm => binary(Value::div, arg(0)?, imm_as_ctrl_ty()?)?,
-        Opcode::UremImm => binary_unsigned(Value::rem, arg(0)?, imm())?,
-        Opcode::SremImm => binary(Value::rem, arg(0)?, imm_as_ctrl_ty()?)?,
+        Opcode::UdivImm => binary_unsigned_can_trap(Value::div, arg(0)?, imm())?,
+        Opcode::SdivImm => binary_can_trap(Value::div, arg(0)?, imm_as_ctrl_ty()?)?,
+        Opcode::UremImm => binary_unsigned_can_trap(Value::rem, arg(0)?, imm())?,
+        Opcode::SremImm => binary_can_trap(Value::rem, arg(0)?, imm_as_ctrl_ty()?)?,
         Opcode::IrsubImm => binary(Value::sub, imm_as_ctrl_ty()?, arg(0)?)?,
         Opcode::IaddCin => unimplemented!("IaddCin"),
         Opcode::IaddIfcin => unimplemented!("IaddIfcin"),
@@ -530,6 +581,7 @@ where
             arg(0)?,
             ValueConversionKind::Truncate(ctrl_ty),
         )?),
+        Opcode::Uunarrow => unimplemented!("Uunarrow"),
         Opcode::Uextend => assign(Value::convert(
             arg(0)?,
             ValueConversionKind::ZeroExtend(ctrl_ty),
@@ -566,14 +618,19 @@ where
         Opcode::FcvtLowFromSint => unimplemented!("FcvtLowFromSint"),
         Opcode::FvpromoteLow => unimplemented!("FvpromoteLow"),
         Opcode::Fvdemote => unimplemented!("Fvdemote"),
-        Opcode::Isplit => unimplemented!("Isplit"),
-        Opcode::Iconcat => unimplemented!("Iconcat"),
+        Opcode::Isplit => assign_multiple(&[
+            Value::convert(arg(0)?, ValueConversionKind::Truncate(types::I64))?,
+            Value::convert(arg(0)?, ValueConversionKind::ExtractUpper(types::I64))?,
+        ]),
+        Opcode::Iconcat => assign(Value::concat(arg(0)?, arg(1)?)?),
         Opcode::AtomicRmw => unimplemented!("AtomicRmw"),
         Opcode::AtomicCas => unimplemented!("AtomicCas"),
         Opcode::AtomicLoad => unimplemented!("AtomicLoad"),
         Opcode::AtomicStore => unimplemented!("AtomicStore"),
         Opcode::Fence => unimplemented!("Fence"),
         Opcode::WideningPairwiseDotProductS => unimplemented!("WideningPairwiseDotProductS"),
+        Opcode::SqmulRoundSat => unimplemented!("SqmulRoundSat"),
+        Opcode::IaddPairwise => unimplemented!("IaddPairwise"),
 
         // TODO: these instructions should be removed once the new backend makes these obsolete
         // (see https://github.com/bytecodealliance/wasmtime/issues/1936); additionally, the
@@ -613,6 +670,9 @@ where
         | Opcode::X86Palignr
         | Opcode::X86ElfTlsGetAddr
         | Opcode::X86MachoTlsGetAddr => unimplemented!("x86 instruction: {}", inst.opcode()),
+        Opcode::JumpTableBase | Opcode::JumpTableEntry | Opcode::IndirectJumpTableBr => {
+            unimplemented!("Legacy instruction: {}", inst.opcode())
+        }
     })
 }
 
@@ -663,7 +723,7 @@ impl<'a, V> ControlFlow<'a, V> {
     }
 }
 
-#[derive(Error, Debug)]
+#[derive(Error, Debug, PartialEq)]
 pub enum CraneliftTrap {
     #[error("user code: {0}")]
     User(TrapCode),
@@ -701,8 +761,8 @@ where
             &left.clone().convert(ValueConversionKind::ToUnsigned)?,
             &right.clone().convert(ValueConversionKind::ToUnsigned)?,
         )?,
-        IntCC::Overflow => unimplemented!("IntCC::Overflow"),
-        IntCC::NotOverflow => unimplemented!("IntCC::NotOverflow"),
+        IntCC::Overflow => Value::overflow(left, right)?,
+        IntCC::NotOverflow => !Value::overflow(left, right)?,
     })
 }
 
